/* For pthreads it may be important on some implementations to ensure
   that pthread.h is included before anything else */
#include <pthread.h>

#include <stdio.h>
#include <stdlib.h>
#include "workers.h"

#define MAX_NUM_WORKERS 16

static int numWorkers = 0;
static int numJobs = 0;
static void (*work)(int i, void *data);
static void *data;
static void (*signal_error)(const char *msg) = NULL;

/* Forward declarations for implementation-dependent routines */
static void releaseWorkers(int njobs, void (*workfun)(int i, void *data),
			   void *workdata);
static void waitForWorkers(void);
static int joinWorkers(void);
static void exit_work(int i, void *data);
static int startWorkerThreads(int n);


/* Error handling routine. Must not return normally. */
static void error(const char *msg)
{
    if (signal_error != NULL)
	signal_error(msg);
    else fprintf(stderr, "%s\n", msg);
    abort();
}


/* The implementations of the following public functions should be
   independent of the particular thread system used. */

void initWorkers(int n, void (*err)(const char *msg))
{
    signal_error = err;

    if (n < 1) error("too few workers");
    if (n > MAX_NUM_WORKERS) error("too many workers");

    if (numWorkers > 0) {
	if (numWorkers == n) return;
	else stopWorkers();
    }
    numWorkers = startWorkerThreads(n);
    if (numWorkers != n) {
	numWorkers = 0;
	/**** shut down any running threads at this point? */
	error("one of the thread creations failed");
    }
}

void runWorkers(int njobs, void (*workfun)(int i, void *data), void *workdata)
{
    if (numWorkers < 1) error("workers have not been initialized");
	
    if (numWorkers == 1) {
	int k;
	for (k = 0; k < njobs; k++)
	    workfun(k, workdata);
    }
    else {
	releaseWorkers(njobs, workfun, workdata);
	waitForWorkers();
    }
}

void stopWorkers()
{
    int join_failed = 0;

    releaseWorkers(numWorkers, exit_work, NULL);
    join_failed = joinWorkers();
    numWorkers = 0;
    if (join_failed) error("a thread join failed");
}


/* The following data structures and routines are pthread-dependent */
/**** Should do more checking of pthread_foo status values. */

typedef struct { int index, running; } worker_info_t;

static pthread_mutex_t work_mtx[MAX_NUM_WORKERS];
static pthread_cond_t wait_cond[MAX_NUM_WORKERS];
static pthread_cond_t done_cond[MAX_NUM_WORKERS];
static pthread_t thread_id[MAX_NUM_WORKERS];
static worker_info_t worker_info[MAX_NUM_WORKERS];

static void *thread_routine(void *arg)
{
    worker_info_t *info = (worker_info_t *) arg;
    int index = info->index;

    pthread_mutex_lock(&work_mtx[index]);
    while (1) {
	worker_info[index].running = 0;
	pthread_cond_signal(&done_cond[index]);
	while (! info->running)
	    pthread_cond_wait(&wait_cond[index], &work_mtx[index]);
	work(index, data);
    }
    return NULL;
}

static void releaseWorkers(int njobs, void (*workfun)(int i, void *data),
			   void *workdata)
{
    int i;

    if (njobs > numWorkers)
	error("non enough workers available");

    numJobs = njobs;
    work = workfun;
    data = workdata;

    for (i = 0; i < njobs; i++) {
	worker_info[i].running = 1;
	pthread_mutex_unlock(&work_mtx[i]);
	pthread_cond_signal(&wait_cond[i]);
    }
}

#define waitForWorker(i) \
    do { pthread_mutex_lock(&work_mtx[i]); \
         while (worker_info[i].running) \
             pthread_cond_wait(&done_cond[i], &work_mtx[i]); \
    } while (0)

static void waitForWorkers()
{
    int i;
    int njobs = numJobs;
    for (i = 0; i < njobs; i++)
	waitForWorker(i);
    numJobs = 0;
}

static int joinWorkers()
{
    int i, status, join_failed = 0;
    void *thread_result;
    for (i = 0; i < numJobs; i++) {
	status = pthread_join(thread_id[i], &thread_result);
	if (status != 0) join_failed = 1;
	pthread_mutex_destroy(&work_mtx[i]);
	pthread_cond_destroy(&wait_cond[i]);
	pthread_cond_destroy(&done_cond[i]);
	worker_info[i].running = 0;
    }
    numJobs = 0;
    return join_failed;
}

static void exit_work(int i, void *data)
{
    pthread_mutex_unlock(&work_mtx[i]);
    pthread_exit(NULL);
}

static int startWorkerThreads(int n)
{
    int i, numStarted, status;

    numStarted = 0;
    for (i = 0; i < n; i++) {
	worker_info[i].index = i;
	worker_info[i].running = 1;
	pthread_mutex_init(&work_mtx[i], NULL);
	pthread_cond_init(&wait_cond[i], NULL);
	pthread_cond_init(&done_cond[i], NULL);
	status = pthread_create(&thread_id[i], NULL, thread_routine,
				&worker_info[i]);
	if (status != 0) break;
	numStarted++;
	waitForWorker(i);
    }
    return numStarted;
}
